from typing import Dict
from urllib.parse import urlparse

from broadcaster._backends.kafka import KafkaBackend as BroadcasterKafkaBackend


class KafkaBackend(BroadcasterKafkaBackend):
    def __init__(self, url: str, bindings: Dict[str, str] = {}):
        super().__init__(url)
        self._servers = urlparse(url).netloc.split(',')

    async def unsubscribe(self, channel: str) -> None:
        self._consumer.unsubscribe()



import asyncio
import functools
from logging import Logger, getLogger
from typing import Any, Awaitable, Dict, Optional
from urllib.parse import urlparse

from broadcaster._backends.base import BroadcastBackend
from google.cloud import pubsub_v1
from google.cloud.pubsub_v1.types import PullResponse, ReceivedMessage

from .. import Event


class GCloudPubSubBackend(BroadcastBackend):
    def __init__(
        self,
        url: str,
        bindings: Dict[str, str] = {},
        logger: Logger = getLogger(__name__),
    ):
        url_parsed = urlparse(url, scheme='gcloud-pubsub')
        self._project = url_parsed.netloc
        self._consumer_channels: Dict[str, str] = {}
        self._producer_channels: Dict[str, str] = {}
        self._channel_index = 0
        self._should_stop = True
        self._logger = logger
        self._set_consumer_config(bindings)

    async def connect(self) -> None:
        self._producer = pubsub_v1.PublisherClient()
        self._consumer = pubsub_v1.SubscriberClient()
        self._should_stop = False

    async def disconnect(self) -> None:
        self._should_stop = True
        self._producer.stop()
        self._consumer.close()
        del self._producer
        del self._consumer

    async def subscribe(self, channel: str) -> None:
        pubsub_channel = self._consumer.subscription_path(
            self._project, channel
        )
        self._consumer_channels[channel] = pubsub_channel

    async def unsubscribe(self, channel: str) -> None:
        self._consumer_channels.pop(channel)

    async def publish(self, channel: str, message: Any) -> None:
        producer_channel = self._producer_channels.get(channel)

        if not producer_channel:
            producer_channel = self._producer.topic_path(
                self._project, channel
            )
            self._producer_channels[channel] = producer_channel

        future = self._producer.publish(producer_channel, message.encode())
        future.result()

    async def next_published(self) -> Optional[Event]:
        channel_index = self._channel_index

        while not self._should_stop:
            channels = list(self._consumer_channels.items())

            if not channels:
                await asyncio.sleep(self._consumer_wait_time)
                continue

            channel_id, pubsub_channel = channels[channel_index]

            response = await self._pull_message_from_consumer(pubsub_channel)

            if not response.received_messages:
                await asyncio.sleep(self._consumer_wait_time)
                if self._channel_index >= len(channels) - 1:
                    channel_index = self._channel_index = 0
                else:
                    self._channel_index += 1
                    channel_index = self._channel_index

            else:
                received_message = response.received_messages[0]
                event = Event(
                    channel_id, received_message.message.data.decode()
                )

                if self._consumer_ack_messages:
                    await self.wait_ack(received_message, pubsub_channel)
                else:
                    event.context['ack_func'] = functools.partial(
                        self.wait_ack, received_message, pubsub_channel,
                    )

                return event

        return None

    async def _pull_message_from_consumer(
        self, pubsub_channel: str
    ) -> PullResponse:
        return await asyncio.get_running_loop().run_in_executor(
            None,
            functools.partial(
                self._consumer.pull,
                pubsub_channel,
                max_messages=1,
                return_immediately=True,
            ),
        )

    def _make_ack_future(
        self, message: ReceivedMessage, pubsub_channel: str
    ) -> Awaitable[None]:
        return asyncio.get_running_loop().run_in_executor(
            None,
            functools.partial(
                self._consumer.acknowledge, pubsub_channel, [message.ack_id],
            ),
        )

    async def wait_ack(
        self,
        message: ReceivedMessage,
        pubsub_channel: str,
        retries_counter: int = 1,
    ) -> None:
        try:
            ack_future = self._make_ack_future(message, pubsub_channel)
            await asyncio.wait_for(
                ack_future, timeout=self._consumer_ack_timeout
            )
        except asyncio.TimeoutError:
            if retries_counter >= self._consumer_ack_retries:
                self._logger.exception(
                    f'ack timeout; {message.message.data.decode()[:100]}...'
                )
            else:
                await self.wait_ack(
                    message, pubsub_channel, retries_counter + 1
                )

    def _set_consumer_config(self, bindings: Dict[str, str]) -> None:
        consumer_wait_time = 1.0
        consumer_ack_messages = False
        consumer_ack_timeout = 5.0
        consumer_ack_retries = 3

        for config_name, config_value in bindings.items():
            if config_name == 'consumer_wait_time':
                consumer_wait_time = float(config_value)

            elif config_name == 'consumer_ack_messages':
                consumer_ack_messages = config_value in (
                    '1',
                    'true',
                    't',
                    'True',
                    'y',
                    'yes',
                )

            elif config_name == 'consumer_ack_timeout':
                consumer_ack_timeout = float(config_value)

            elif config_name == 'consumer_ack_retries':
                consumer_ack_retries = int(config_value)

        self._consumer_wait_time = consumer_wait_time
        self._consumer_ack_messages = consumer_ack_messages
        self._consumer_ack_timeout = consumer_ack_timeout
        self._consumer_ack_retries = consumer_ack_retries



import os
import signal
import time
from asyncio import AbstractEventLoop, get_event_loop
from typing import Any, Optional

import typer

from asyncapi import (
    AsyncApi,
    UrlOrModuleRequiredError,
    build_api,
    build_api_auto_spec,
)


def main(
    url: Optional[str] = typer.Option(None, envvar='ASYNCAPI_URL'),
    server: Optional[str] = typer.Option(None, envvar='ASYNCAPI_SERVER'),
    server_bindings: Optional[str] = typer.Option(
        None, envvar='ASYNCAPI_SERVER_BINDINGS'
    ),
    api_module: str = typer.Option('', envvar='ASYNCAPI_MODULE'),
    republish_errors: Optional[bool] = typer.Option(
        None, envvar='ASYNCAPI_REPUBLISH_ERRORS'
    ),
    channel: Optional[str] = typer.Option(None, envvar='ASYNCAPI_CHANNEL'),
    workers: int = typer.Option(1, envvar='ASYNCAPI_WORKERS'),
    channels_subscribes: Optional[str] = typer.Option(
        None, envvar='ASYNCAPI_CHANNELS_SUBSCRIBES'
    ),
    republish_errors_channels: Optional[str] = typer.Option(
        None, envvar='ASYNCAPI_REPUBLISH_ERRORS_CHANNELS'
    ),
) -> None:

    if url is None:
        if api_module is None:
            raise UrlOrModuleRequiredError()

        api = build_api_auto_spec(
            module_name=api_module,
            server=server,
            server_bindings=server_bindings,
            republish_errors=republish_errors,
            channels_subscribes=channels_subscribes,
            republish_errors_channels=republish_errors_channels,
        )

    else:
        api = build_api(
            url,
            server=server,
            module_name=api_module,
            server_bindings=server_bindings,
            republish_errors=republish_errors,
            channels_subscribes=channels_subscribes,
            republish_errors_channels=republish_errors_channels,
        )

    fork_app(workers)
    graceful_stop()
    loop = get_event_loop()
    start(loop, api, channel)
    loop.run_forever()


def start(
    loop: AbstractEventLoop, api: AsyncApi, channel: Optional[str]
) -> None:
    async def init() -> None:
        await api.connect()

        if channel is None:
            await api.listen_all()
        else:
            await api.listen(channel)

    task = loop.create_task(init())
    task.add_done_callback(task_callback)

    typer.echo('Waiting messages...')


def task_callback(future: Any) -> None:
    future.result()  # pragma: no cover


def run() -> None:
    typer.run(main)


def graceful_stop() -> None:
    def exit_gracefully(signum: int = -1, frame: Any = None) -> None:
        signame = {  # type: ignore
            signal.SIGINT: 'SIGINT',
            signal.SIGTERM: 'SIGTERM',
        }.get(signum)
        typer.echo(f'Received {signame} scheduling shutdown...')

        time.sleep(1)
        raise SystemExit(0)

    signal.signal(signal.SIGINT, exit_gracefully)
    signal.signal(signal.SIGTERM, exit_gracefully)


def fork_app(workers: int) -> None:
    if workers > 1:
        typer.echo(f'Forking for {workers} workers')
        for _ in range(workers - 1):
            pid = os.fork()
            if pid == 0:
                break


if __name__ == '__main__':
    run()



from typing import Any, Dict
from urllib.parse import urlparse

from broadcaster import Broadcast
from broadcaster._backends.base import BroadcastBackend


class EventsHandler(Broadcast):
    _backend: BroadcastBackend

    def __init__(self, url: str, bindings: Dict[str, Any] = {}):
        super().__init__(url)

        parsed_url = urlparse(url)

        if parsed_url.scheme == 'kafka':
            from .backends.kafka import KafkaBackend

            self._backend = KafkaBackend(url, bindings)

        elif parsed_url.scheme == 'gcloud-pubsub':
            from .backends.gcloud_pubsub import GCloudPubSubBackend

            self._backend = GCloudPubSubBackend(url, bindings)



class AsyncApiError(Exception):
    ...


class ReferenceNotFoundError(AsyncApiError):
    ...


class InvalidChannelError(AsyncApiError):
    ...


class OperationIdNotFoundError(AsyncApiError):
    ...


class ChannelOperationNotFoundError(AsyncApiError):
    ...


class UrlOrModuleRequiredError(AsyncApiError):
    ...


class InvalidJsonSchemaTypeError(AsyncApiError):
    ...


class InvalidMessageError(AsyncApiError):
    ...


class EmptyServersError(AsyncApiError):
    ...


class ServerNotFoundError(AsyncApiError):
    ...


class InvalidContentTypeError(AsyncApiError):
    ...


class InvalidAsyncApiVersionError(AsyncApiError):
    ...


class InvalidServerBindingError(AsyncApiError):
    ...


class InvalidServerBindingProtocolError(AsyncApiError):
    ...


class InvalidChannelsSubscribersError(AsyncApiError):
    ...


class ChannelPublishNotFoundError(AsyncApiError):
    ...



from google.cloud import pubsub_v1
from google.api_core.exceptions import AlreadyExists as AlreadyExistsError


def pubsub_init():
    project_id = 'asyncapi-local'
    topics = ['chatroom', 'chatroom2', 'user-update']
    subscriptions = ['chatroom', 'chatroom2', 'user-update-custom-sub']

    for topic_name, subscription_name in zip(topics, subscriptions):
        topic_path = create_topic(project_id, topic_name)
        create_subscription(project_id, subscription_name, topic_path)


def create_topic(project_id, name):
    publisher = pubsub_v1.PublisherClient()
    topic_path = publisher.topic_path(project_id, name)

    try:
        publisher.create_topic(topic_path)
    except AlreadyExistsError:
        ...

    return topic_path


def create_subscription(project_id, name, topic_path):
    subscriber = pubsub_v1.SubscriberClient()
    subscription_path = subscriber.subscription_path(project_id, name)

    try:
        subscriber.create_subscription(subscription_path, topic_path)
    except AlreadyExistsError:
        ...


pubsub_init()



from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Sequence,
    Type,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)


def type_as_jsonschema(python_type: Type[Any]) -> Dict[str, Any]:
    if python_type is Any:
        return {}

    schema_type = SCALARS.get(python_type)

    if schema_type is None:
        origin = getattr(python_type, '__origin__', None)

        if origin is List or origin is Sequence or origin is Iterable:
            return {
                'type': 'array',
                'items': type_as_jsonschema(python_type.__args__[0]),
            }

        return build_object_schema(python_type)

    return {'type': schema_type}


def build_object_schema(python_type: Type[Any]) -> Dict[str, Any]:
    type_hints = get_type_hints(python_type)
    origin = get_origin(python_type)
    schema: Dict[str, Any] = {
        'type': 'object',
        'properties': {},
    }
    required = []

    for attr_name, attr_type in type_hints.items():
        origin = get_origin(attr_type)
        args = get_args(attr_type)

        if origin is Union and len(args) == 2 and type(None) in args:
            attr_schema = type_as_jsonschema(args[0])

        elif origin is Union:
            if not type(None) in args:
                required.append(attr_name)

            attr_schema = {'anyOf': [type_as_jsonschema(arg) for arg in args]}

        else:
            required.append(attr_name)
            attr_schema = type_as_jsonschema(attr_type)

        schema['properties'][attr_name] = attr_schema

    if required:
        schema['required'] = required

    return schema


SCALARS = {
    int: 'integer',
    str: 'string',
    float: 'number',
    bool: 'boolean',
}



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['asyncapi', 'asyncapi._tests', 'asyncapi.events', 'asyncapi.events.backends']

package_data = \
{'': ['*'], 'asyncapi.events.backends': ['_tests/*']}

install_requires = \
['broadcaster', 'jsondaora']

extras_require = \
{'dev-doc': ['mkdocs', 'mkdocs-material', 'markdown-include'],
 'docs': ['apidaora'],
 'gcloud-pubsub': ['google-cloud-pubsub>=1,<2'],
 'http': ['requests'],
 'kafka': ['aiokafka'],
 'postgres': ['asyncpg'],
 'redis': ['asyncio-redis'],
 'subscriber': ['typer'],
 'test': ['black',
          'isort',
          'ipython',
          'mypy',
          'pytest-asyncio',
          'pytest-cov',
          'pytest-mock',
          'pytest',
          'flake8',
          'bumpversion',
          'jsondaora',
          'asynctest'],
 'yaml': ['pyyaml']}

entry_points = \
{'console_scripts': ['asyncapi-docs = asyncapi:run_docs',
                     'asyncapi-subscriber = asyncapi:run_subscriber']}

setup(name='asyncapi',
      version='0.11.0',
      description='asyncapi',
      author='Diogo Dutra',
      author_email='diogodutradamata@gmail.com',
      url='https://github.com/dutradda/asyncapi',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.8',
     )
